#!/bin/bash
#
# Script to collect a report about Flashphoner WebCallServer work
#


# Global variables configuration

# Declare global variables
function declareGlobals() {
  PRODUCT=FlashphonerWebCallServer
  WCS_HOME="/usr/local/$PRODUCT"
  FLASHPHONER_USER="flashphoner"
  REPORT_LOG_FILE="$WCS_HOME/logs/report.log"
  LAST_HOURS=3
  REPORTS_HOME="$WCS_HOME/report"
  REPORT_DATE=`date +%Y-%m-%d-%H-%M-%S`
  REPORT_DIR="$REPORTS_HOME/report-$REPORT_DATE"
  WCS_PID=""
}

# Get product name
function getProduct() {
  echo "$PRODUCT"
}

# Get WCS home directory
function getHome() {
  echo "$WCS_HOME"
}

# Get WCS user name
function getUser() {
  echo "$FLASHPHONER_USER"
}

# Get report log file name
function getReportLogFile() {
  echo "$REPORT_LOG_FILE"
}

# Get last hours to collect logs
function getLastHours() {
  echo "$LAST_HOURS"
}

# Set last hours to collect logs
function setLastHours() {
  local lastHours=$1

  [[ ! -z $lastHours ]] && LAST_HOURS=$lastHours
}

# Get report date
function getReportDate() {
  echo "$REPORT_DATE"
}

# Set report date
function setReportDate() {
  local reportDate=${1:-$(date +%Y-%m-%d-%H-%M-%S)}

  [[ ! -z $reportDate ]] && REPORT_DATE=$reportDate
}

# Get report directory
function getReportDir() {
  echo "$REPORT_DIR"
}

# Set report directory
function setReportDir() {
  local reportDir=${1:-$(getReportsHome)/report-$(getReportDate)}

  [[ ! -z $reportDir ]] && REPORT_DIR=$reportDir
}

# Get WCS pid
function getWcsPid() {
  echo "$WCS_PID"
}

# Set WCS pid
function setWcsPid() {
  local pid=$1

  [[ ! -z $pid ]] && WCS_PID=$pid
}

# Get reports home dir
function getReportsHome() {
  echo "$REPORTS_HOME"
}


# Logging functions

# Declare logging variables
function declareLog() {
  local logFile=$1

  DEBUG=${DEBUG:-false}
  DEFAULT_LOG=${DEFAULT_LOG:-}

  setDefaultLogFile $logFile
}

# Check if debug log enabled
function isDebugEnabled() {
  if [[ ! -z $DEBUG ]]; then
    $DEBUG; return
  else
    false; return
  fi
}

# Enable debug logging
function enableDebug() {
  DEBUG=true
}

# Disable debug logging
function disableDebug() {
  DEBUG=false
}

# Get default log file
function getDefaultLogFile() {
  echo "$DEFAULT_LOG"
}

# Set default log file
function setDefaultLogFile() {
  local logFile=$1

  [[ -z $logFile ]] && return 1
  if checkLogPath $logFile; then
    DEFAULT_LOG=$logFile
  fi
}

# # Check if log file path exists and make the dir if needed
function checkLogPath() {
  local logFile=$1
  local result=0

  [[ -z $logFile ]] && return 1
  logPath=$(dirname "$logFile")
  if [[ ! -d $logPath ]]; then
    mkdir -p $logPath
    result=$?
  fi
  return $result
}

# Form log timestamp
function logDate() {
  local dateToLog=`echo "[$(date '+%Y-%m-%d %H:%M:%S')]"`
 
  echo "$dateToLog"
}

# Write to a log file
function writeLog() {
  local level=$1
  local message=$2
  local logFile=$3
  local owner=$(getCurrentUser)
  local logPath=""
  local callerFuncName=""

  [[ -z $logFile ]] && logFile=$(getDefaultLogFile)
  if [[ "$level" == "" || "$message" == "" || "$logFile" == "" ]]; then
    return 1
  fi
  if [[ -f $logFile ]]; then
    owner=$(ls -l $logFile | awk '{print $3}');
  else
    checkLogPath $logFile
    [[ $? -ne 0 ]] && return 1
  fi
  # Determine caller function name
  for callerFuncName in ${FUNCNAME[@]}; do
    if [[ "$callerFuncName" == "${FUNCNAME[0]}" || "$callerFuncName" =~ ^log ]]; then
      continue
    else
      break
    fi
  done
  if [[ $owner != $(getCurrentUser) ]]; then
    # The log file is owned by someone else, try to sudo to it
    sudo -u $owner echo -e "$(logDate) $level $callerFuncName - $message" >> $logFile
  else
    echo -e "$(logDate) $level $callerFuncName - $message" >> $logFile
  fi
 }

# Write INFO level message
function logInfo() {
  local message=$1
  local logFile=$2

  writeLog "INFO" "$message" "$logFile"
}

# Write WARN level message
function logWarn() {
  local message=$1
  local logFile=$2

  writeLog "WARN" "$message" "$logFile"
}

# Write ERROR level message
function logError() {
  local message=$1
  local logFile=$2

  writeLog "ERROR" "$message" "$logFile"
}

# Write DEBUG level message
function logDebug() {
  local message=$1
  local logFile=$2

  if isDebugEnabled; then
    writeLog "DEBUG" "$message" "$logFile"
  fi
}


# Various helper functions

# Check if user exists in system
function isUser() {
  local user=$1
  
  if [[ -z $user ]]; then
    user=$(getUser)
  fi
  if [[ "$(getent passwd $user)" =~ ^$user ]]; then
    true; return
  else
    false; return
  fi
}

# Get current user running script
function getCurrentUser() {
  echo "$(whoami)"
}

# Get daemon type
function getDaemonType() {
  local daemonType=`ps --no-headers -o comm 1`
  echo "$daemonType"
} 

# Check if service is running
function isService() {
  # Check service only if systemd used #WCS-3034
  if [[ $(getDaemonType) != 'systemd' ]]; then
    false; return
  fi
  if systemctl is-active --quiet webcallserver.service; then
    true; return
  else
    false; return
  fi
}

# Check if the script is running from root
function isScriptRunningFromRoot() {
  if [[ "$(getCurrentUser)" == "root" ]]; then
    true; return
  else
    false; return
  fi
}

# Set file system item owner to user
function setOwner() {
  local user=$1
  local object=$2

  if [[ "$user" == "" || "$object" == "" ]]; then
    return 1
  fi
  [[ ! isScriptRunningFromRoot ]] && return 1
  [[ ! -e $object ]] && return 1
  chown -RHL $user:$(id -gn $user) $object
}

# Get Java version
function getJavaVersion() {
  local jdkVersion=`java -version 2>&1`
  local javaFullVersion=`echo $jdkVersion | head -1 | cut -d" " -f 3 | tr -d \"`
  local javaMajor=`echo $javaFullVersion | cut -d \. -f 1`
  local javaMinor=0
  local javaVersion=0

  if [[ $javaMajor -eq 1 ]]; then
    javaMinor=`echo $javaFullVersion | cut -d \. -f 2`
    if [[ $javaMinor -ge 8  ]]; then
      javaVersion=$javaMinor
    fi
  elif [[ $javaMajor -ge 8 ]]; then
    javaVersion=$javaMajor
  fi
  logDebug "Java version $javaVersion"
  echo "$javaVersion"
}

# Get user running the process by pid
function getUserByPid() {
  local pid=$1
  local user=`ps -o uname= -p "$pid"`
 
  logDebug "Process $pid running as user $user"
  echo "$user"
}

# Check if user is owner of the object
function isOwned() {
  local object=$1
  local user=$2
  local owner=""
  
  if [[ -z $user ]]; then
    user=$(getUser)
  fi
  owner=$(stat -c "%U" $object)
  if [[ $owner == $user ]]; then
    true; return
  else
    false; return
  fi
}

# Command detection functions

# Return full command by name or emty string if not exist
function getCommand() {
  local shortCmd=$1
  local fullCmd=""
  
  if [[ ! -z $shortCmd ]]; then
    fullCmd=`command -v $shortCmd 2>/dev/null`
  fi
  echo "$fullCmd"
}

# Check if command exists
function isCommandExists() {
  local shortCmd=$1
  local fullCmd=$(getCommand $shortCmd)

  if [[ -z $fullCmd ]]; then
    echo "'$shortCmd' not found"
    logError "Command '$shortCmd' not found"
    false; return
  fi
  true; return
}


# Server logs gathering functions

# Collect logs
function get_logs() {
  local dateFrom=`date "+%Y-%m-%d %H" -d "$(getLastHours) hour ago"`
  local dateTo=`date "+%Y-%m-%d %H" -d "1 hour"`

  mkdir -p $(getReportDir)/logs/server_logs
  mkdir -p $(getReportDir)/logs/client_logs
  cd $(getHome)/logs
  find . -newermt "$dateFrom" ! -newermt "$dateTo" -exec cp --parents -t $(getReportDir)/logs {} + > /dev/null 2>&1
  return 0
}



# Server network interface statistics gathering functions

# Collect netstat output
function get_netstat() {
  local cmd="netstat"
  local fullCmd=""

  ! isCommandExists $cmd && return 1
  fullCmd=$(getCommand $cmd)
  $fullCmd -antpul > "$(getReportDir)/$cmd.log"
  return $?
}



# Opened objects statistics gathering functions

# Collect lsof output
function get_lsof() {
  local cmd="lsof"
  local fullCmd=""

  ! isCommandExists $cmd && return 1
  fullCmd=$(getCommand $cmd)
  $fullCmd -P -p $(getWcsPid) > "$(getReportDir)/lsof.log"
  return $?
}



# Memory statistics gathering functions

# Detect pmap keys
function detectPmapKeys() {
  local cmd=$1

  if $cmd -h | grep -e "^[[:space:]]\-X" > /dev/null; then
    cmd="$cmd -X"
  else
    cmd="$cmd -x"
  fi

  echo "$cmd"
}

# Collect pmap output
function get_pmap() {
  local cmd="pmap"
  local fullCmd=""

  ! isCommandExists $cmd && return 1
  fullCmd=$(detectPmapKeys $(getCommand $cmd))
  $fullCmd $(getWcsPid) > "$(getReportDir)/pmap.log"
  return $?
}



# Jstack gathering functions

# Run jstack as user depending on WCS running from
function getJstackAsUser() {
  local user=$1
  local cmd=$2
  local pid=$3
  local javaVersion=$(getJavaVersion)

  if [[ "$user" == "root" ]]; then
    $cmd $pid > "$(getReportDir)/jstack.log"
    return $?
  fi
  if [[ $javaVersion -gt 8 && isService ]]; then
    # Run jstack from root for service
    getJstackAsUser root $fullCmd $pid
  else
    sudo -u $user $cmd $pid > "$(getReportDir)/jstack.log"
  fi
  return $?
}

# Collect jstack output
function get_jstack() {
  local cmd="jstack"
  local fullCmd=""
  local pid=$(getWcsPid)
  local user=$(getUserByPid $pid)

  ! isCommandExists $cmd && return 1
  fullCmd=$(getCommand $cmd)
  getJstackAsUser $user $fullCmd $pid
  return $?
}



# Heap dump gathering functions

# Run jcmd as user depending on WCS running from
function runGcAsUser() {
  local user=$1
  local cmd=$2
  local pid=$3
  local javaVersion=$(getJavaVersion)

  if [[ "$user" == "root" ]]; then
    $cmd $pid GC.run > /dev/null 2>&1
    return $?
  fi
  if [[ $javaVersion -gt 8 && isService ]]; then
    # Run jcmd from root for service
    runGcAsUser root $fullCmd $pid
  else
    sudo -u $user $cmd $pid GC.run > /dev/null 2>&1
  fi
  return $?
}

# Run jmap as user depending on WCS running from
function runJmapAsUser() {
  local user=$1
  local cmd=$2
  local pid=$3
  local javaVersion=$(getJavaVersion)

  if [[ "$user" == "root" ]]; then
    $cmd -dump:format=b,file=$(getReportDir)/dump/$pid.hprof $pid > /dev/null 2>&1
    return $?
  fi
  if [[ $javaVersion -gt 8 && isService ]]; then
    # Run jmap from root for service
    runJmapAsUser root $fullCmd $pid
  else
    sudo -u $user $cmd -dump:format=b,file=$(getReportDir)/dump/$pid.hprof $pid > /dev/null 2>&1
  fi
  return $?
}

# Run GC using jcmd
function runGc() {
  local cmd="jcmd"
  local fullCmd=""
  local pid=$(getWcsPid)
  local user=$(getUserByPid $pid)

  ! isCommandExists $cmd && return 1
  fullCmd=$(getCommand $cmd)
  runGcAsUser $user $fullCmd $pid
  return $?
}

# Collect heap dump using jmap
function getHeapDump() {
  local cmd="jmap"
  local fullCmd=""
  local pid=$(getWcsPid)
  local user=$(getUserByPid $pid)

  ! isCommandExists $cmd && return 1
  fullCmd=$(getCommand $cmd)
  mkdir -p $(getReportDir)/dump
  isUser && setOwner $(getUser) $(getReportDir)/dump
  runJmapAsUser $user $fullCmd $pid
  return $?
}

# Run GC then collect heap dump
function get_dump() {
  runGc
  getHeapDump
  return $?  
}



# Server configs gathering functions

# Collect configs
function get_conf() {
  local confDir=$(getHome)/conf
  local targetDir=$(getReportDir)/conf

  mkdir -p $targetDir
  cp $confDir/flashphoner.properties $targetDir
  cp $confDir/wcs-core.properties $targetDir
  cp $confDir/log4j.properties $targetDir
  cp $confDir/WCS.version $targetDir
  cp $confDir/*.yml $targetDir
  if [[ "$(echo $confDir/*.sdp)" != "$confDir/*.sdp" ]]; then
    cp $confDir/*.sdp $targetDir
  fi
  return 0
}



# Server system information gathering functions

# Collect system info
function get_sysinfo() {
  local targetDir=$(getReportDir)/sysinfo

  mkdir -p $targetDir
  cat /proc/cpuinfo > "$targetDir/cpuinfo.log"
  cat /proc/meminfo > "$targetDir/meminfo.log"
  df -h > "$targetDir/df.log"
  if isCommandExists ifconfig; then
    ifconfig > "$targetDir/ifcfg.log"
  elif isCommandExists ip; then
    ip a > "$targetDir/ifcfg.log"
  fi
  iptables -nvL > "$targetDir/iptables.log"
  return 0
}



# WCS statistics information gathering functions

# Collect stats info
function get_stats() {
  local config="$(getHome)/conf/flashphoner.properties"
  local httpPort=""
  local statsQuery=""
  local res=0

  httpPort=`sed -n -e "s/^http.port[ \t]*=[ \t]*\(.*\)/\1/p" $config`
  [[ -z $httpPort ]] && httpPort="8081"
  statsQuery="http://localhost:$httpPort/?action=stat&format=json"
  logDebug "$statsQuery"
  if isCommandExists jq; then
    curl -s "$statsQuery" | jq . > $(getReportDir)/stats.json
  else
    curl -s "$statsQuery" > $(getReportDir)/stats.json
  fi
  return $?
}



# Report archiving functions

# Pack report to archive
function get_tar() {
  local tarFile="report_$(getReportDate).tar.gz"

  ! isCommandExists tar && return 1
  logInfo "Packing report to archive $tarFile"
  cd $(getReportsHome)
  tar -zcf $tarFile $(basename $(getReportDir)) > /dev/null 2>&1
  return $?
}


# Main module

# Usage displaying
function usage() {
  echo "Usage: $(basename $0) [OPTIONS]"
  echo "Default report will include netstat, lsof, logs between now and 3 hours ago, pmap, jstack"
  echo -e "  --conf\t\t copy configuration"
  echo -e "  --dump\t\t make heap dump"
  echo -e "  --sysinfo\t\t gather system info"
  echo -e "  --stats\t\t gather WCS statistics"
  echo -e "  --tar\t\t\t tar report"
  echo -e "  --hours <hours>\t hours count to collect lates logs (3 by default)"
  echo -e "  --help\t\t help"
  return 1
}

# Check the process and prepare report dir
function prepareToReport() {
  local pidFileRoot=/var/run/FlashphonerMainWebCallServer.pid
  local pidFileNonRoot=$(getHome)/bin/FlashphonerMainWebCallServer.pid
  local pid=""
  local msgMustBeRoot="The script must be run as root"
  local msgNotRunning="Server is not running"

  if [[ "$(getCurrentUser)" != "root" ]]; then
    logError "$msgMustBeRoot"
    echo -e "$msgMustBeRoot:\n sudo report.sh"
    false; return
  fi

  if [ -f $pidFileRoot ]; then
    pid=$(cat $pidFileRoot)
  elif [ -f $pidFileNonRoot ]; then
    pid=$(cat $pidFileNonRoot)
  else
    logWarn "Server is probably not running, looking for the process"
    pid=$(pgrep -fn com.flashphoner.server.Server)
  fi
  if [[ -z $pid ]]; then
    echo "$msgNotRunning"
    logError "$msgNotRunning"
    false; return
  fi

  setWcsPid $pid
  setReportDate `date +%Y-%m-%d-%H-%M-%S`
  setReportDir
  mkdir -p "$(getReportDir)"
  if isUser $(getUser); then
    setOwner $(getUser) $(getReportDir)/
  fi
  logInfo "Collecting report $(getReportDir)"
}

# Display scheduled actions
function showSchedule() {
  local report=""
  local dumpSize=""
  local coreConfig="$(getHome)/conf/wcs-core.properties"

  logInfo "Scheduled report:"
  echo -e "Scheduled report:\n"
  for report in $@; do
    if [[ "$report" == "dump" ]]; then
      dumpSize=$(grep -e "^-Xmx" $coreConfig | cut -f2 -d"x")
      [[ -z $dumpSize ]] && dumpSize=$(grep -Eo "[[:space:]]-Xmx[a-zA-Z0-9\/]*" $coreConfig | cut -d" " -f 2 | cut -f2 -d"x")
      echo "* $report [estimated $dumpSize]"
      logInfo "Task: dump, estimated dump size: $dumpSize"
    else
      logInfo "Task: $report"
      echo "* $report"
    fi
  done
}

# Pack report to tar.gz
function packReport() {
  local tarFile="report_$(getReportDate).tar.gz"

  logInfo "Packing report to archive $tarFile"
  cd $(getReportsHome)
  tar -zcf $tarFile $(basename $(getReportDir)) > /dev/null 2>&1
  return $?
}

# Main script function
function main() {
  local startTime=$(date +%s)
  local scheduledReport=(
    logs
    netstat
    lsof
    pmap
    jstack
  )
  local tarReport=false
  local taskResult=0

  declareGlobals
  declareLog $(getReportLogFile)
  # Set log file owner
  if isUser $(getUser) && isScriptRunningFromRoot; then
    touch $(getReportLogFile)
    setOwner $(getUser) $(getReportLogFile)
  fi

  # Parse command line
  while [[ $# -gt 0 ]]; do
    case "$1" in
      --help)
        usage
        return 1
        ;;
      --dump)
        scheduledReport+=(dump);;
      --sysinfo)
        scheduledReport+=(sysinfo);;
      --conf)
        scheduledReport+=(conf);;
      --stats)
        scheduledReport+=(stats);;
      --tar)
        tarReport=true;;
      --hours)
        shift
        setLastHours $1;;
      --debug)
        enableDebug;;
      *)
        usage
        return 1
        ;;
    esac
    shift 
  done
  if ! prepareToReport; then
    return 1
  fi
  if $tarReport; then
    scheduledReport+=(tar)
  fi
  showSchedule ${scheduledReport[@]}
  echo -e "\nProgress: \n"
  for task in ${scheduledReport[@]}; do
    echo "[PROGRESS] $task"
    tput cuu1
    if get_$task; then
      tput el
      echo "[DONE] $task"
    else
      taskResult=$?
      tput el
      echo "[FAILED] $task"
      logError "Task $task failed, result $taskResult"
    fi
  done
  endTime=$(date +%s)
  echo "Report complete in $((endTime - startTime)) seconds. Check $(getReportDir)"
  logInfo "Report $(getReportDir) collected"
  if isUser $(getUser); then
    setOwner $(getUser) $(getReportsHome)/
    setOwner $(getUser) $(getReportLogFile)
  fi
}

main "$@"

exit $?


